from typing import Union, List

from fastapi import Security


class RouteConfig:
    guards: List[Security]

    def __init__(self, guards: List[Security] = None):
        self.guards = []
        if guards is not None:
            self.guards = guards


class ModelConfig:
    get_one: Union[RouteConfig, bool]
    batch_get: Union[RouteConfig, bool]
    get_link_one: Union[RouteConfig, bool]
    get_all: Union[RouteConfig, bool]
    get_link_all: Union[RouteConfig, bool]
    create_one: Union[RouteConfig, bool]
    update_one: Union[RouteConfig, bool]
    delete_one: Union[RouteConfig, bool]

    def __init__(self):
        self.get_one = RouteConfig()
        self.batch_get = RouteConfig()
        self.get_link_one = RouteConfig()
        self.get_all = RouteConfig()
        self.get_link_all = RouteConfig()
        self.create_one = RouteConfig()
        self.update_one = RouteConfig()
        self.delete_one = RouteConfig()

    def add_guard(self, guard):
        self.get_one.guards.append(guard)
        self.get_link_one.guards.append(guard)
        self.get_all.guards.append(guard)
        self.get_link_all.guards.append(guard)
        self.create_one.guards.append(guard)
        self.update_one.guards.append(guard)
        self.delete_one.guards.append(guard)



import unittest
import datetime

from typing import Optional, List
from pydantic import BaseModel

from fastapi_toolkit.generate.field_helper import FieldHelper, FieldType, Link


class A(BaseModel):
    a: int


class B:
    a: int


class TesterFieldHelper(unittest.TestCase):
    def setUp(self):
        self.fh = FieldHelper()

    def test_builtin(self):
        self.assertEqual(True, self.fh.is_builtin(str))
        self.assertEqual(True, self.fh.is_builtin(int))
        self.assertEqual(True, self.fh.is_builtin(float))
        self.assertEqual(True, self.fh.is_builtin(bool))
        self.assertEqual(True, self.fh.is_builtin(bytes))
        self.assertEqual(False, self.fh.is_builtin(A))
        self.assertEqual(False, self.fh.is_builtin(B))

    def test_model(self):
        self.assertEqual(True, self.fh.is_model(A))
        self.assertEqual(False, self.fh.is_model(B))
        self.assertEqual(False, self.fh.is_model(int))

    def test_optional(self):
        self.assertEqual(True, self.fh.is_optional(Optional[A]))
        self.assertEqual(False, self.fh.is_optional(A))
        self.assertEqual(True, self.fh.is_optional(Optional[int]))
        self.assertEqual(False, self.fh.is_optional(int))

    def test_batch(self):
        self.assertEqual(True, self.fh.is_batch(List[A]))
        self.assertEqual(True, self.fh.is_batch(List[int]))
        self.assertEqual(True, self.fh.is_batch(List[B]))
        self.assertEqual(False, self.fh.is_batch(int))
        self.assertEqual(False, self.fh.is_batch(A))
        self.assertEqual(False, self.fh.is_batch(B))

    def test_explicit_parse(self):
        tests = [
            (FieldType(python_type='str', sql_type='sqltypes.Text'), self.fh.parse_builtin(str)),
            (FieldType(python_type='int', sql_type='sqltypes.Integer'), self.fh.parse_builtin(int)),
            (FieldType(python_type='float', sql_type='sqltypes.Float'), self.fh.parse_builtin(float)),
            (FieldType(python_type='bool', sql_type='sqltypes.Boolean'), self.fh.parse_builtin(bool)),
            (FieldType(python_type='bytes', sql_type='sqltypes.LargeBinary'), self.fh.parse_builtin(bytes)),

            (FieldType(python_type='A', link=Link(model="A", type="one")), self.fh.parse_model(A)),
            (None, self.fh.parse_model(int)),
            (None, self.fh.parse_model(B)),

            (FieldType(python_type='Optional[A]', link=Link(model="A", type="one"), nullable=True),
             self.fh.parse_optional(Optional[A])),
            (None, self.fh.parse_optional(Optional[B])),

            (FieldType(python_type='List[A]', link=Link(model="A"), type="many"), self.fh.parse_batch(List[A])),
            (None, self.fh.parse_batch(List[B])),
            (None, self.fh.parse_batch(List[int])),
        ]
        for want, got in tests:
            self.assertEqual(want, got)

    def test_parse(self):
        tests = [
            (FieldType(python_type='str', sql_type='sqltypes.Text'), str),
            (FieldType(python_type='int', sql_type='sqltypes.Integer'), int),
            (FieldType(python_type='float', sql_type='sqltypes.Float'), float),
            (FieldType(python_type='bool', sql_type='sqltypes.Boolean'), bool),
            (FieldType(python_type='bytes', sql_type='sqltypes.LargeBinary'), bytes),

            (FieldType(python_type='datetime.datetime', sql_type='sqltypes.DateTime', depends=['datetime']),
             datetime.datetime),
            (FieldType(python_type='datetime.date', sql_type='sqltypes.Date', depends=['datetime']),
             datetime.date),

            (FieldType(python_type='A', link=Link(model="A", type="one")), A),
            (None, B),

            (FieldType(python_type='Optional[A]', link=Link(model="A", type="one"), nullable=True), Optional[A]),
            (None, Optional[B]),

            (FieldType(python_type='List[A]', link=Link(model="A", type="many"), nullable=True), List[A]),
            (None, List[B]),
        ]

        for want, got in tests:
            self.assertEqual(want, self.fh.parse(got))


if __name__ == '__main__':
    unittest.main()



from typing import Type
from abc import ABC, abstractmethod
from fastapi import APIRouter
from fastapi_pagination import Page


class BaseRouter(APIRouter, ABC):
    snake_name: str
    snake_plural_name: str
    camel_name: str
    base_schema: Type
    schema: Type

    def __init__(self, config, base_schema: Type, schema: Type):
        super().__init__()
        self.prefix = f"/{self.snake_name}"
        self.tags = [self.camel_name]
        self.base_schema = base_schema
        self.schema = schema

        if config.get_one:
            self.add_api_route(
                path="/get_one",
                endpoint=self._get_one(),
                response_model=self.base_schema,
                methods=["POST"],
                dependencies=config.get_one.guards,
                summary=f"Get one {self.snake_name}",
                description=f"Get one {self.snake_name}",
                response_description=f"A {self.snake_name}",
            )

        if config.batch_get:
            self.add_api_route(
                path="/batch_get",
                endpoint=self._batch_get(),
                response_model=Page[self.base_schema],
                methods=["POST"],
                dependencies=config.get_one.guards,
                summary=f"Batch get {self.snake_name}",
                description=f"Batch get {self.snake_name}",
                response_description=f"Batch {self.snake_name}",
            )

        if config.get_all:
            self.add_api_route(
                path="/get_all",
                endpoint=self._get_all(),
                response_model=Page[self.base_schema],
                methods=["POST"],
                dependencies=config.get_all.guards,
                summary=f"Get all {self.snake_name}",
                description=f"Get all {self.snake_name}",
                response_description=f"All {self.snake_name}",
            )

        if config.get_link_all:
            self.add_api_route(
                path="/get_link_all",
                endpoint=self._get_link_all(),
                response_model=Page[self.schema],
                methods=["POST"],
                dependencies=config.get_link_all.guards,
                summary=f"Get all {self.snake_plural_name} with link",
                description=f"Get all {self.snake_plural_name} with link",
                response_description=f"All {self.snake_plural_name} with link",
            )

            if config.create_one:
                self.add_api_route(
                    path="/create_one",
                    endpoint=self._create_one(),
                    response_model=self.base_schema,
                    methods=["POST"],
                    dependencies=config.create_one.guards,
                    summary=f"Create one {self.snake_name}",
                    description=f"Create one {self.snake_name}",
                    response_description=f"Created {self.snake_name}",
                )

            if config.update_one:
                self.add_api_route(
                    path="/update_one",
                    endpoint=self._update_one(),
                    response_model=self.base_schema,
                    methods=["POST"],
                    dependencies=config.update_one.guards,
                    summary=f"Update one {self.snake_name}",
                    description=f"Update one {self.snake_name}",
                    response_description=f"Updated {self.snake_name}",
                )

            if config.delete_one:
                self.add_api_route(
                    path="/delete_one",
                    endpoint=self._delete_one(),
                    methods=["POST"],
                    dependencies=config.delete_one.guards,
                    summary=f"Delete one {self.snake_name}",
                    description=f"Delete one {self.snake_name}",
                    response_description=f"Deleted {self.snake_name}",
                )

    @abstractmethod
    def _get_one(self):
        raise NotImplementedError

    @abstractmethod
    def _batch_get(self):
        raise NotImplementedError

    @abstractmethod
    def _get_all(self):
        raise NotImplementedError

    @abstractmethod
    def _get_link_all(self):
        raise NotImplementedError

    @abstractmethod
    def _create_one(self):
        raise NotImplementedError

    @abstractmethod
    def _update_one(self):
        raise NotImplementedError

    @abstractmethod
    def _delete_one(self):
        raise NotImplementedError



from typing import Type, List, Optional, Callable

from fastapi import FastAPI, APIRouter
from fastapi_pagination import Page

from fastapi_toolkit.define import Schema


class Endpoint:
    def __init__(self, repo_map):
        self.repo_map = repo_map

    def _get_repo(self, s: Type[Schema]):
        def dfs(a: Type[Schema], p: List[Type[Schema]]):
            for i in a.__subclasses__():
                yield from dfs(i, p.copy() + [a])
                if i is not s:
                    continue
                yield p + [a, i]

        try:
            path = next(dfs(Schema, []))
        except StopIteration:
            raise ValueError(f"can't find repo for {s.__name__}")

        for c in path:
            if c in self.repo_map:
                return self.repo_map[c]

    def get_one(self, router: FastAPI | APIRouter, url: str, func: Optional[Callable[..., Schema]] = None):
        def decorator(view: Type[Schema]):
            nonlocal func
            if func is None:
                func = self._get_repo(view).get_one
            router.add_api_route(
                path=url,
                endpoint=func,
                methods=['POST'],
                tags=[view.__name__],
                response_model=view,
            )

            return view

        return decorator

    def get_many(self, router: FastAPI | APIRouter, url: str, func: Optional[Callable[..., Page[Schema]]] = None):
        def decorator(view: Type[Schema]):
            nonlocal func
            if func is None:
                func = self._get_repo(view).get_all
            router.add_api_route(
                path=url,
                endpoint=self._get_repo(view).get_all,
                methods=['POST'],
                tags=[view.__name__],
                response_model=Page[view],
            )

            return view

        return decorator



import datetime
from enum import Enum
from typing import Type, Union, Optional, Set
from pydantic import BaseModel, Field as PField
from pydantic.fields import FieldInfo, PydanticUndefined


class LinkType(str, Enum):
    one = "one"
    many = "many"


class Link(BaseModel):
    model: str
    type: LinkType


class FieldType(BaseModel):
    python_type: str
    sql_type: str = PField(default="")
    nullable: bool = PField(default=False)
    link: Optional[Link] = PField(default=None)
    depends: Set[str] = PField(default_factory=set)
    default: Optional[str] = PField(default=None)


class FieldHelper:

    @staticmethod
    def is_builtin(t: Type):
        return t in [str, int, float, bool, bytes, datetime.datetime, datetime.date]

    def parse_builtin(self, t: Type):
        if not self.is_builtin(t):
            return
        t: [str, int, float, bool, bytes]
        sql_map = {
            str: "sqltypes.Text",
            int: "sqltypes.Integer",
            float: "sqltypes.Float",
            bool: "sqltypes.Boolean",
            bytes: "sqltypes.LargeBinary",
            datetime.datetime: "sqltypes.DateTime",
            datetime.date: "sqltypes.Date",
        }
        f = FieldType(python_type=str(t.__name__), sql_type=sql_map[t])
        if t.__module__ != "builtins":
            f.python_type = f"{t.__module__}.{f.python_type}"
            f.depends.add(t.__module__)
        return f

    def is_model(self, t: Type):
        if self.is_builtin(t):
            return False
        return isinstance(t, type) and BaseModel.__subclasscheck__(t)

    def parse_model(self, t: Type):
        if not self.is_model(t):
            return
        return FieldType(python_type=t.__name__, link=Link(model=t.__name__, type="one"))

    def is_optional(self, t: Type):
        if self.is_builtin(t):
            return False
        if not hasattr(t, "__origin__"):
            return False
        origin = getattr(t, "__origin__")
        if origin is not Union:
            return False
        if not hasattr(t, "__args__"):
            return False
        args = getattr(t, "__args__")
        if len(args) != 2:
            return False
        if args[1] is not type(None):
            return False
        return True

    def parse_optional(self, t: Type):
        if not self.is_optional(t):
            return
        if self.is_builtin(t.__args__[0]):
            f = self.parse_builtin(t.__args__[0])
        elif self.is_model(t.__args__[0]):
            f = self.parse_model(t.__args__[0])
        else:
            return None
        f.nullable = True
        f.python_type = f'Optional[{f.python_type}]'
        return f

    def is_batch(self, t: Type):
        if self.is_builtin(t):
            return False
        if not hasattr(t, "__origin__"):
            return False
        origin = getattr(t, "__origin__")
        if origin is not list:
            return False
        return True

    def parse_batch(self, t: Type):
        if not self.is_batch(t):
            return None
        if self.is_model(t.__args__[0]):
            f = self.parse_model(t.__args__[0])
            f.sql_type = ""
            f.python_type = f'List[{f.python_type}]'
            f.nullable = True
            f.link.type = "many"
            return f

    def parse(self, f: FieldInfo):
        t = f.annotation
        res = self.parse_type(t)
        if self.is_builtin(t):
            if f.default is not PydanticUndefined:
                res.default = f.default
        return res

    def parse_type(self, t: Type):
        f = self.parse_builtin(t)
        if f is not None:
            return f
        f = self.parse_model(t)
        if f is not None:
            return f
        f = self.parse_optional(t)
        if f is not None:
            return f
        f = self.parse_batch(t)
        if f is not None:
            return f



import unittest
from fastapi_toolkit.configer import Configer


class ConfigerTest(unittest.TestCase):
    def test_init(self):
        configer = Configer('test_configer.yaml')
        self.assertIsInstance(configer, Configer)

    def test_get_value(self):
        configer = Configer('test_configer.yaml')
        self.assertEqual(configer['key'], 'value')

    def test_get_inner_value(self):
        configer = Configer('test_configer.yaml')
        self.assertEqual(configer['keyA.keyB.keyC'], 'inner_value')

    def test_get_none(self):
        configer = Configer('test_configer.yaml')
        self.assertEqual(configer['key.none'], None)

    def test_get_inner_dict(self):
        configer = Configer('test_configer.yaml')
        self.assertEqual(configer['arr.2'], {'k': 'v'})

    def test_get_list_value(self):
        configer = Configer('test_configer.yaml')
        self.assertEqual(configer['arr.0'], 1)

    def test_get_list_overflow(self):
        configer = Configer('test_configer.yaml')
        self.assertEqual(configer['arr.3'], None)

    def test_get_inner_list_value(self):
        configer = Configer('test_configer.yaml')
        self.assertEqual(configer['keyA.keyB.arr'], [1, 2, 3])



import datetime
import inspect
import os
import hashlib
from collections import defaultdict

import typer
from typing import Callable, Any, Sequence, Dict, List, Optional, Tuple, Literal, Type
from jinja2 import Environment, PackageLoader
# import networkx as nx

from pydantic import BaseModel, Field as PField
from fastapi_toolkit.define import Schema, Controller
from .field_helper import FieldHelper, FieldType, LinkType

from .sql_mapping import mapping
from .utils import to_snake, plural

GENERATE_FUNC = Callable[[], str]


class NameInfo(BaseModel):
    origin: str
    snake: str
    snake_plural: str
    camel: str
    camel_plural: str
    table: str
    db: str
    schema: str
    base_schema: str
    fk: str


class FK(BaseModel):
    name: str
    key_type: str = "int"
    column: str


class AssociationTable(BaseModel):
    name: str
    left: 'ModelRenderData'
    right: 'ModelRenderData'


class LinkRenderData(BaseModel):
    link_name: str
    target_type: str
    back_populates: str

    schema_link_name: str
    schema_target_type: str
    default_value: str = ''


class Link(BaseModel):
    link_name: str
    alias: Optional[str] = None
    origin: "ModelRenderData"
    target: "ModelRenderData"
    fk: Optional[FK] = None
    table: Optional[AssociationTable] = None
    type: LinkType

    render_data: Optional[LinkRenderData] = None
    link_op_names: List[str] = PField(default_factory=list)
    link_op_codes: List[str] = PField(default_factory=list)

    def make_render(self, pair_link: "Link"):
        self.render_data = LinkRenderData(link_name=self.link_name, target_type=f'"{self.target.name.db}"',
                                          back_populates=pair_link.link_name,
                                          schema_link_name=self.link_name,
                                          schema_target_type=self.target.name.base_schema)
        if self.type == LinkType.many:
            self.render_data.target_type = f'List[{self.render_data.target_type}]'
            self.render_data.schema_target_type = f'List[{self.render_data.schema_target_type}]'
            self.render_data.default_value = 'Field(default_factory=list)'
        else:
            self.render_data.target_type = f'Optional[{self.render_data.target_type}]'
            self.render_data.schema_target_type = f'Optional[{self.render_data.schema_target_type}]'
            self.render_data.default_value = 'None'

    def link_prefix(self):
        if self.type == "one":
            target_name = self.target.name.snake
        else:
            target_name = self.target.name.snake_plural
        l_name = self.link_name
        if self.alias is not None:
            l_name = self.alias
        if l_name[-len(target_name):] != target_name:
            raise ValueError(f"link name {l_name} not match target {self.target.name.origin}")
        return l_name[:-len(target_name)]


class Field(BaseModel):
    name: NameInfo
    alias: Optional[NameInfo] = None
    type: FieldType


class ModelRenderData(BaseModel):
    name: NameInfo
    model: Type[Schema] = None
    fields: List[Field] = []
    links: List[Link] = []


class MethodRenderData(BaseModel):
    name: str
    defines: List[str]
    args: Dict[str, str]
    res_type: str
    process: List[str]
    res_fields: Dict[str, str]


class ControllerRenderData(BaseModel):
    name: NameInfo
    methods: List[MethodRenderData]


class CodeGenerator:
    def __init__(self, root_path='inner_code'):
        self.root_path = root_path
        self.models_path = os.path.join(root_path, 'models.py')
        self.schemas_path = os.path.join(root_path, 'schemas.py')
        self.dev_path = os.path.join(root_path, 'dev')
        self.crud_path = os.path.join(root_path, 'repo')
        self.routers_path = os.path.join(root_path, 'routers')
        self.auth_path = os.path.join(root_path, 'auth')
        self.api_path = os.path.join(root_path, 'api')
        self.stub_path = os.path.join(root_path, 'stub')

        self.force_rewrite = False
        self.show_network = False
        self.async_repo = False

        if not os.path.exists(self.root_path):
            os.mkdir(self.root_path)
        if not os.path.exists(self.dev_path):
            os.mkdir(self.dev_path)
        if not os.path.exists(self.crud_path):
            os.mkdir(self.crud_path)
        if not os.path.exists(self.routers_path):
            os.mkdir(self.routers_path)
        if not os.path.exists(self.auth_path):
            os.mkdir(self.auth_path)
        if not os.path.exists(self.api_path):
            os.mkdir(self.api_path)
        if not os.path.exists(self.stub_path):
            os.mkdir(self.stub_path)
        self.env = Environment(
            loader=PackageLoader('fastapi_toolkit', 'templates'),
            trim_blocks=True, lstrip_blocks=True)

        self.define_schemas: Dict[str, Type[Schema]] = {}
        self.pydantic_schemas: Dict[str, Type[Schema]] = {}
        self.links: List[Tuple[Literal["one"] | Literal["many"], Type[Schema], Type[Schema]]] = []
        self.custom_types: List[Dict[str, Any]] = []
        # self.model_network = nx.Graph()
        self.association_tables: List[AssociationTable] = []
        self.crds: List[ControllerRenderData] = []
        self.parse()

    @staticmethod
    def _check_file_valid(path):
        with open(path, 'r') as f:
            line = f.readline()
            if not line.startswith('# generate_hash:'):
                return False
            content_hash = line.split(':')[1].strip()
            f.readline()
            f.readline()
            f.readline()
            content = f.read()
            return hashlib.md5(content.encode('utf8')).hexdigest() == content_hash

    def _generate_file(self, path, func: GENERATE_FUNC):
        content = func()
        generate_hash = hashlib.md5(content.encode('utf8')).hexdigest()
        if os.path.exists(path):
            with open(path, 'r') as f:
                line = f.readline()
                if line.startswith('# generate_hash:'):
                    head_hash = line.split(':')[1].strip()
                else:
                    head_hash = None
                if head_hash == generate_hash and CodeGenerator._check_file_valid(path):
                    print(f'file {path} is up to date, skip generate')
                    return
                elif not self.force_rewrite:
                    overwrite = typer.confirm(
                        f'file {path} has been changed or is out of date, do you want to overwrite it?')
                    if not overwrite:
                        return
        with open(path, 'w') as f:
            f.write(f'# generate_hash: {generate_hash}\n')
            f.write(f'"""\n'
                    f'This file was automatically generated in {datetime.datetime.now()}\n'
                    f'"""\n')
            f.write(content)

    def parse(self):
        self._parse_models()
        # self._parse_network()
        self._parse_api()
        # self._parse_mock()

    def _get_schemas(self, root=Schema):
        for model_ in root.__subclasses__():
            yield model_
            yield from self._get_schemas(model_)

    def _parse_models(self):
        self.define_schemas = {schema.__name__: schema for schema in self._get_schemas()}
        self.model_render_data = {
            schema_name: self._make_render_data_field(schema)
            for schema_name, schema in self.define_schemas.items()
        }
        self.build_links()

    def build_links(self):
        links: Dict[str, List[Link]] = defaultdict(list)
        for model in self.model_render_data.values():
            for field in filter(lambda x: x.type.link is not None, model.fields):
                field: Field
                link = Link(link_name=field.name.origin, origin=model,
                            target=self.model_render_data[field.type.link.model], type=field.type.link.type)
                if field.alias is not None:
                    link.alias = field.alias.origin
                links[model.name.origin].append(link)
            model.fields = list(filter(lambda x: x.type.link is None, model.fields))
            model.fields.sort(key=lambda x: x.type.nullable or x.type.default != None)

        link_groups: List[Tuple[Link, Link]] = []
        visited_link = set()
        for ls in links.values():
            for link in ls:
                if id(link) in visited_link:
                    continue
                target_links = links[link.target.name.origin]
                for target_link in target_links:
                    if id(target_link) in visited_link:
                        continue
                    if (link.link_prefix() == target_link.link_prefix()
                            and link.target.name.origin == target_link.origin.name.origin
                            and target_link.target.name.origin == link.origin.name.origin):
                        visited_link.add(id(link))
                        visited_link.add(id(target_link))
                        link_groups.append((link, target_link))
                        break
                else:
                    raise ValueError(f"unable to pair {link.origin.name.origin}'s link "
                                     f"{link.link_name} to target {link.target.name.origin}")

        def make_one_many_link(l_one: Link, l_many: Link):
            l_one.fk = FK(name=f'_fk_{l_one.link_name}_{l_one.target.name.table}_id',
                          column=f"{l_one.target.name.table}.id")
            l_one.origin.links.append(l_one)
            l_many.origin.links.append(l_many)
            l_one.make_render(l_many)
            l_many.make_render(l_one)

        for l1, l2 in link_groups:
            t1, t2 = l1.type, l2.type
            match (t1, t2):
                case (LinkType.one, LinkType.one):
                    l1.origin.links.append(l1)
                    l2.origin.links.append(l2)
                    l1.fk = FK(name=f'_fk_{l1.link_name}_{l1.target.name.table}_id',
                               column=f'{l1.target.name.table}.id')
                    l1.make_render(l2)
                    l2.make_render(l1)
                case (LinkType.one, LinkType.many):
                    make_one_many_link(l1, l2)
                case (LinkType.many, LinkType.one):
                    make_one_many_link(l2, l1)
                case (LinkType.many, LinkType.many):
                    at_name = ['association_table']
                    if l1.link_prefix() != "":
                        at_name.append(l1.link_prefix().replace('_', ''))
                    at_name += [l1.origin.name.table, l2.origin.name.table]
                    association_table = AssociationTable(
                        name='_'.join(at_name),
                        left=l1.origin, right=l2.origin
                    )
                    l1.table = association_table
                    l2.table = association_table
                    self.association_tables.append(association_table)
                    l1.origin.links.append(l1)
                    l2.origin.links.append(l2)
                    l1.make_render(l2)
                    l2.make_render(l1)

        def build_link_op(link: Link, name: str, from_id: bool, from_batch: bool):
            arg = link.target.name.snake
            arg_type = link.target.name.base_schema
            if from_id:
                arg += '_id'
                arg_type = 'int'
            if from_batch:
                arg += 's'
                arg_type = f'List[{arg_type}]'
            filter_expr = f'__eq__({arg})'
            if from_batch:
                if from_id:
                    filter_expr = f'in_({arg})'
                else:
                    filter_expr = f'in_(map(lambda x: x.id, {arg}))'
            elif not from_id:
                filter_expr = f'__eq__({arg}.id)'
            query_code = f'query = query.join({link.target.name.db}).filter({link.target.name.db}.id.{filter_expr})'
            o = link.origin
            for l in o.links:
                op = 'selectinload' if l.type is LinkType.many else 'joinedload'
                query_code += f'\n    query = query.options({op}({o.name.db}.{l.link_name}))'
            link.link_op_codes.append(
                f"""
def {name}_query({arg}: {arg_type}, query=Depends(get_all_query)) -> Select:
    if type(query) is not Select:
        query = get_all_query(QueryParams())
    {query_code}
    return query
                """
            )
            link.link_op_codes.append(
                f"""
{'async ' if self.async_repo else ''}def {name}({arg}: {arg_type}, db=Depends(get_db), query=Depends(get_all_query)) -> List[{link.origin.name.db}]:
    if type(query) is not Select:
        query = get_all_query(QueryParams())
    {query_code}
    return {'(await db.scalars(query)).all() ' if self.async_repo else 'db.scalars(query).all()'}
                """
            )

        for link in (l for m in self.model_render_data.values() for l in m.links):
            build_link_op(link, f'get_{link.link_name}_id_is', True, False)
            build_link_op(link, f'get_{link.link_name}_is', False, False)
            build_link_op(link, f'get_{link.link_name}_id_has', True, True)
            build_link_op(link, f'get_{link.link_name}_has', False, True)

    def _make_render_data_field(self, schema: Type[Schema]):
        fh = FieldHelper()
        fields = []
        for name, field in schema.model_fields.items():
            fields.append(Field(name=self._name_info(name), alias=self._name_info(field.alias), type=fh.parse(field)))
        m = ModelRenderData(
            name=self._name_info(schema.__name__),
            fields=fields
        )
        if m.name.origin == 'User':
            m.fields.append(Field(name=self._name_info('user_key'), type=fh.parse_type(str)))
        return m

    def _parse_api(self):

        def make_render(name, args, res) -> MethodRenderData:
            fh = FieldHelper()
            defines = []

            def f(v):
                t = v.annotation
                if fh.is_builtin(t):
                    return fh.parse(v).python_type
                if fh.is_model(t):
                    if fh.parse_model(t).python_type not in self.model_render_data:
                        c = f'class {fh.parse(t).python_type}(BaseModel):\n'
                        t: Type[BaseModel]
                        for k, v in t.model_fields.items():
                            c += f'        {k}: {f(v.annotation)}\n'
                        defines.append(c)
                        return fh.parse(t).python_type
                    return self.model_render_data[fh.parse(t).python_type].name.base_schema
                if fh.is_batch(t):
                    py_t = fh.parse_batch(t).python_type
                    arg = py_t[5:-1]
                    if arg in self.model_render_data:
                        py_t = f'List[{self.model_render_data[arg].name.base_schema}]'
                    return py_t
                return fh.parse(t).python_type

            process = []
            inputs = {}
            # 解析需要输入什么
            for k, v in args.items():
                if fh.is_builtin(v):
                    inputs[k] = f(v)
                    continue
                if fh.is_model(v):
                    inputs[f'{k}_ident'] = 'int'
                    crud_name = self.model_render_data[v.__name__].name.snake + '_crud'
                    process.append(f'{k} = await {crud_name}.get_one({k}_ident, db)')
                    continue
                else:
                    raise ValueError(f'Unsupported type {v}')
            # 解析需要输出什么
            res: BaseModel
            res_fields = {}
            for k, v in res.model_fields.items():
                res_fields[k] = f(v)
            return MethodRenderData(
                name=name,
                args=inputs,
                defines=defines,
                res_type=f(res),
                process=process,
                res_fields=res_fields,
            )

        def get_controller(root=Controller):
            for model_ in root.__subclasses__():
                yield model_
                yield from get_controller(model_)

        for c in get_controller():
            crd = ControllerRenderData(name=self._name_info(c.__name__), methods=[])
            for n, f in inspect.getmembers(c, predicate=inspect.isfunction):
                args = {}
                res = type(None)
                for k, v in f.__annotations__.items():
                    if k == 'return':
                        res = v
                        continue
                    args[k] = v
                crd.methods.append(make_render(n, args, res))
            self.crds.append(crd)

    # def _parse_network(self):
    #     self.model_network.add_nodes_from(self.model_render_data.keys())
    #     for model in self.model_render_data.values():
    #         for link in model.links:
    #             self.model_network.add_edge(link.origin.name.origin, link.target.name.origin)
    #
    #     pos = nx.spring_layout(self.model_network)
    #     edge_labels = {
    #         (link.origin.name.origin, link.target.name.origin): link.link_name
    #         for model in self.model_render_data.values() for link in model.links
    #     }
    #     if self.show_network:
    #         nx.draw(self.model_network.to_directed(), pos, with_labels=True)
    #         nx.draw_networkx_edge_labels(
    #             self.model_network.to_directed(), pos,
    #             edge_labels=edge_labels, label_pos=0.75
    #         )
    #         import matplotlib.pyplot as plt
    #         plt.show()

    def _parse_mock(self, export=False):
        self.model_network.add_nodes_from(self.model_render_data.keys())
        self.mock_dependency = {}
        self.mock_root = []
        for component_nodes in nx.connected_components(self.model_network):
            subgraph = self.model_network.subgraph(component_nodes)
            mst = nx.minimum_spanning_tree(subgraph)
            max_degree_node = max(subgraph.degree, key=lambda x: x[1])[0]
            rooted_tree = nx.dfs_tree(mst, source=max_degree_node)
            if export:
                import matplotlib.pyplot as plt
                pos = nx.planar_layout(self.model_network)
                nx.draw(rooted_tree.to_directed(), pos, with_labels=True)
                plt.show()
            self.mock_root.append(max_degree_node)
            for node in rooted_tree.nodes():
                self.mock_dependency[node] = list(rooted_tree.successors(node))

    @staticmethod
    def _name_info(name: str) -> Optional[NameInfo]:
        if type(name) is not str:
            return None
        return NameInfo(
            origin=name,
            snake=to_snake(name),
            snake_plural=plural(to_snake(name)),
            camel=name,
            camel_plural=plural(name),
            table=to_snake(name),
            db=f'DB{name}',
            schema=f'Schema{name}',
            base_schema=f'SchemaBase{name}',
            fk=f'fk_{to_snake(name)}_id',
        )

    def _define2table(self) -> str:
        template = self.env.get_template('models/main.py.jinja2')
        return template.render(
            deps=self.custom_types,
            models=self.model_render_data.values(),
            association_tables=self.association_tables,
        )

    def _define2schema(self) -> str:
        template = self.env.get_template('schemas/main.py.jinja2')
        return template.render(
            deps=self.custom_types,
            models=self.model_render_data.values(),
        )

    def _from_template(self, template_name: str, **kwargs):
        def func():
            return self.env.get_template(template_name).render(**kwargs)

        return func

    def _generate_tables(self, auth_type):
        self._generate_file(os.path.join(self.root_path, 'db.py'), self._from_template('db.py.jinja2'))
        self._generate_file(os.path.join(self.root_path, 'setting.py'),
                            self._from_template('setting.py.jinja2', auth_type=auth_type))
        self._generate_file(self.models_path, self._define2table)
        self._generate_file(self.schemas_path, self._define2schema)
        self._generate_file(
            os.path.join(self.dev_path, 'db.py'),
            self._from_template(
                'dev.db.py.jinja2',
                root_path=str(self.root_path).replace('/', '.').replace('\\', '.')))
        self._generate_file(os.path.join(self.dev_path, '__init__.py'), lambda: '')

    def _generate_routers(self):
        template_path = 'async/' if self.async_repo else 'sync/'
        models = self.model_render_data.values()
        for model in models:
            self._generate_file(os.path.join(self.crud_path, f'{model.name.snake}_repo.py'),
                                self._from_template(f'repo/{template_path}main.py.jinja2', model=model))
            self._generate_file(os.path.join(self.routers_path, f'{model.name.snake}_router.py'),
                                self._from_template('routers/main.py.j2', model=model))
            self._generate_file(os.path.join(self.stub_path, f'{model.name.snake}_stub.py'),
                                self._from_template('stub/main.py.jinja2', model=model))
        self._generate_file(os.path.join(self.crud_path, '__init__.py'),
                            self._from_template(f'repo/{template_path}__init__.py.jinja2',
                                                models=models))
        self._generate_file(os.path.join(self.routers_path, '__init__.py'), self._from_template(
            'routers/init.py.j2',
            models=models,
        ))
        self._generate_file(os.path.join(self.stub_path, '__init__.py'), self._from_template(
            'stub/__init__.py.jinja2',
            models=models,
        ))

    def _generate_mock(self):
        self._generate_file(
            os.path.join(self.root_path, 'mock.py'), self._from_template(
                'mock.py.j2',
                deps=self.mock_dependency,
                mock_root=self.mock_root,
                models=self.model_render_data.values()))

    def _generate_auth(self, mode: str):
        user_model = self.model_render_data['User']
        self._generate_file(os.path.join(self.auth_path, '__init__.py'),
                            self._from_template(f'auth/{mode}/__init__.py.j2'))
        if mode != 'key':
            self._generate_file(os.path.join(self.auth_path, 'models.py'),
                                self._from_template(f'auth/{mode}/models.py.j2', model=user_model))
        self._generate_file(os.path.join(self.auth_path, 'routes.py'), self._from_template(f'auth/{mode}/routes.py.j2'))

    def _generate_config(self):
        self._generate_file(os.path.join(self.root_path, 'config.py'), self._from_template(
            'config.py.j2', models=self.model_render_data.values()))

    def _generate_custom_types(self):
        self._generate_file(os.path.join(self.root_path, 'custom_types.py'), self._from_template(
            'custom_types.py.j2', custom_types=self.custom_types))

    def _generate_apis(self, name):
        for crd in self.crds:
            if name != '' and crd.name.origin.lower() != name.lower():
                continue
            path = os.path.join(self.api_path, f'{crd.name.snake}.py')
            while os.path.exists(path):
                path = path + '.new'
            self._generate_file(path, self._from_template(
                'api/main.py.j2', crd=crd,
                cruds=map(lambda x: x.name.snake + '_crud', self.model_render_data.values())))

    def generate(self, table: bool = True, router: bool = True, mock: bool = True, auth: str = ""):
        self._generate_custom_types()
        if table:
            self._generate_tables(auth_type=auth)
        if router:
            self._generate_routers()
        # if mock:
        #     self._generate_mock()
        if auth != "":
            self._generate_auth(mode=auth)
        # self._generate_apis()
        self._generate_config()

    def generate_api(self, name):
        self._generate_apis(name)



from typing import Union, Any
from pathlib import Path

import yaml


class Configer:
    def __init__(self, config: Union[Path, str] = Path('fastapi_toolkit.yaml')):
        if isinstance(config, str):
            path = Path(config)
        else:
            path = config
        if path.is_file():
            self.config = yaml.load(path.read_text(), Loader=yaml.FullLoader)
        else:
            self.config = {}

    def __getitem__(self, key: str) -> Union[Any, None]:
        v = self.config
        for i in key.split('.'):
            if v is None:
                return None
            if isinstance(v, list):
                i = int(i)
                if i >= len(v):
                    return None
                v = v[i] or None
            elif isinstance(v, dict):
                v = v.get(i, None)
            else:
                return None
        return v

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __delitem__(self, key):
        raise NotImplementedError



import datetime
import enum
import uuid
from typing import Type, Tuple, Optional

import pydantic
from sqlalchemy.sql import sqltypes


def mapping(t) -> Tuple[bool, Optional[str]]:
    type_map = {
        int: sqltypes.Integer,
        float: sqltypes.Float,
        str: sqltypes.Text,
        bool: sqltypes.Boolean,
        bytes: sqltypes.LargeBinary,
        bytearray: sqltypes.LargeBinary,

        datetime.datetime: sqltypes.DateTime,
        datetime.date: sqltypes.Date,
        datetime.timedelta: sqltypes.Interval,
    }
    if t in type_map:
        return False, 'sqltypes.' + type_map[t].__name__
    if issubclass(t, enum.Enum):
        # TODO: check database is postgresql
        return True, f'dialects.postgresql.ENUM({t.__name__})'
    return True, None



from abc import ABC, abstractmethod
import uuid
from datetime import timedelta, datetime
from typing import Union, Optional, List, Protocol

from fastapi import Depends, HTTPException, status
from fastapi.security import OAuth2PasswordBearer, SecurityScopes

from jose import jwt, JWTError
from passlib.context import CryptContext
from pydantic import BaseModel


class UserProtocol(Protocol):
    id: uuid.UUID
    username: str
    scopes: Optional[str]


class UserCreateProtocol(Protocol):
    username: str
    password: str


class UserFullProtocol(UserProtocol):
    is_superuser: bool
    is_active: bool
    hashed_password: str

    registered_at: datetime
    activated_at: Optional[datetime]
    last_login_at: Optional[datetime]
    accessed_at: Optional[datetime]


class AuthDBBackend(ABC):

    @abstractmethod
    async def get_user(self, username: str) -> Optional[UserFullProtocol]:
        raise NotImplementedError

    @abstractmethod
    async def add_user(self, create: UserCreateProtocol, hashed_password: str) -> UserFullProtocol:
        raise NotImplementedError

    @abstractmethod
    async def access(self, username: str) -> None:
        raise NotImplementedError


credentials_exception = HTTPException(
    status_code=status.HTTP_401_UNAUTHORIZED,
    detail="Could not validate credentials",
    headers={"WWW-Authenticate": "Bearer"},
)


class Token(BaseModel):
    access_token: str
    token_type: str


class TokenData(BaseModel):
    username: Union[str, None] = None


class Auth:

    def __init__(self, secret_key, algorithm, access_token_expire_minutes, db_backend: AuthDBBackend,
                 schema_user, schema_user_full):
        self.pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
        self.oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")
        self.SECRET_KEY, self.ALGORITHM = secret_key, algorithm
        self.ACCESS_TOKEN_EXPIRE_MINUTES = access_token_expire_minutes
        self.backend = db_backend
        self.SchemaUser = schema_user
        self.SchemaUserFull = schema_user_full

    def verify_password(self, plain_password: str, hashed_password: str) -> bool:
        return self.pwd_context.verify(plain_password, hashed_password)

    def get_password_hash(self, password: str) -> str:
        return self.pwd_context.hash(password)

    def create_access_token(self, data: dict) -> str:
        to_encode = data.copy()
        expire = datetime.utcnow() + timedelta(minutes=self.ACCESS_TOKEN_EXPIRE_MINUTES)
        to_encode.update({"exp": expire})
        encoded_jwt = jwt.encode(to_encode, self.SECRET_KEY, algorithm=self.ALGORITHM)
        return encoded_jwt

    async def authenticate_user(self, username: str, password: str):
        user = await self.backend.get_user(username)
        if not user:
            return False
        if not self.verify_password(password, user.hashed_password):
            return False
        return user

    def require_user(self, is_activate: bool = False, is_superuser: bool = False):
        SchemaUserFull = self.SchemaUserFull
        SchemaUser = self.SchemaUser

        async def _get_current_user(token: str = Depends(self.oauth2_scheme)) -> SchemaUserFull:
            try:
                payload = jwt.decode(token, self.SECRET_KEY, algorithms=[self.ALGORITHM])
                username: str = payload.get("sub")
                if username is None:
                    raise credentials_exception
                token_data = TokenData(username=username)
            except JWTError:
                raise credentials_exception
            user = await self.backend.get_user(token_data.username)
            await self.backend.access(token_data.username)
            return SchemaUserFull.model_validate(user)

        async def func(security_scopes: SecurityScopes, user=Depends(_get_current_user)) -> SchemaUser:
            if is_activate and not user.is_active:
                raise credentials_exception

            if is_superuser and not user.is_superuser:
                raise credentials_exception

            if security_scopes.scopes:
                if not user.scopes:
                    raise credentials_exception
                if not all(scope in user.scopes for scope in security_scopes.scopes):
                    raise credentials_exception

            return SchemaUser.model_validate(user)

        return func



from typing import Callable, Generator, Any, Annotated, Optional, Union, Coroutine, TypeVar

from sqlalchemy.orm import Session
from typing_extensions import Doc


def computed_field(db_func: Optional[Callable[..., Generator[Session, Any, None]]] = None):
    def decorator(func, *args, **kw):
        from pydantic import computed_field

        if db_func is not None:
            @computed_field(*args, **kw)
            def wrapper(*args_, **kw_) -> func.__annotations__.get('return'):
                return func(*args_, **kw_, db=next(db_func()))
        else:
            @computed_field(*args, **kw)
            def wrapper(*args_, **kw_) -> func.__annotations__.get('return'):
                return func(*args_, **kw_)
        return wrapper

    return decorator


T = TypeVar("T")


def Depends(  # noqa: N802
        dependency: Annotated[
            Optional[
                Union[
                    Callable[..., Coroutine[Any, Any, T]],
                    Callable[..., Generator[T, Any, Any]],
                    Callable[..., T],
                ]
            ],
            Doc(
                """
                A "dependable" callable (like a function).

                Don't call it directly, FastAPI will call it for you, just pass the object
                directly.
                """
            ),
        ] = None,
        *,
        use_cache: Annotated[
            bool,
            Doc(
                """
                By default, after a dependency is called the first time in a request, if
                the dependency is declared again for the rest of the request (for example
                if the dependency is needed by several dependencies), the value will be
                re-used for the rest of the request.

                Set `use_cache` to `False` to disable this behavior and ensure the
                dependency is called again (if declared more than once) in the same request.
                """
            ),
        ] = True,
) -> T:
    from fastapi import Depends
    return Depends(dependency, use_cache=use_cache)



import inspect

from fastapi_toolkit.generate import CodeGenerator
from fastapi_toolkit.define import Controller


class ApiGenerator:
    g: CodeGenerator

    def __init__(self, g: CodeGenerator):
        self.g = g
        for c in self._get_controller():
            for name, f in inspect.getmembers(c, predicate=inspect.isfunction):
                print(name, f.__annotations__['return'])
                r = f.__annotations__['return']

    def _get_controller(self, root=Controller):
        for model_ in root.__subclasses__():
            yield model_
            yield from self._get_controller(model_)



from abc import ABC, abstractmethod
import uuid
from datetime import timedelta, datetime
from typing import Union, Optional, List, Protocol

from fastapi import Depends, HTTPException, status
from fastapi.security import APIKeyHeader

from jose import jwt, JWTError
from passlib.context import CryptContext
from pydantic import BaseModel


class UserProtocol(Protocol):
    id: uuid.UUID
    user_key: str


class UserCreateProtocol(Protocol):
    user_key: str


class UserFullProtocol(UserProtocol):
    is_superuser: bool
    is_active: bool
    hashed_password: str

    registered_at: datetime
    activated_at: Optional[datetime]
    last_login_at: Optional[datetime]
    accessed_at: Optional[datetime]


class AuthDBBackend(ABC):

    @abstractmethod
    def get_user(self, user_key: str) -> Optional[UserFullProtocol]:
        raise NotImplementedError

    @abstractmethod
    def add_user(self, create: UserCreateProtocol) -> UserFullProtocol:
        raise NotImplementedError

    @abstractmethod
    def access(self, user_key: str) -> None:
        raise NotImplementedError


credentials_exception = HTTPException(
    status_code=status.HTTP_401_UNAUTHORIZED,
    detail="Could not validate credentials",
    headers={"WWW-Authenticate": "Bearer"},
)


class Token(BaseModel):
    access_token: str
    token_type: str


class TokenData(BaseModel):
    username: Union[str, None] = None


class Auth:

    def __init__(self, db_backend: AuthDBBackend, key_name:str, schema_user, schema_user_full):
        self.auth_scheme = APIKeyHeader(name=key_name)
        self.backend = db_backend
        self.SchemaUser = schema_user
        self.SchemaUserFull = schema_user_full

    def authenticate_user(self, username: str, password: str):
        user = self.backend.get_user(username)
        if not user:
            return False
        return user

    def require_user(self, is_activate: bool = False, is_superuser: bool = False):
        SchemaUserFull = self.SchemaUserFull
        SchemaUser = self.SchemaUser

        def _get_current_user(token: str = Depends(self.auth_scheme)) -> SchemaUserFull:
            user = self.backend.get_user(token)
            if user is None:
                raise credentials_exception
            return SchemaUserFull.model_validate(user)

        def func(user=Depends(_get_current_user)) -> SchemaUser:
            if is_activate and not user.is_active:
                raise credentials_exception

            if is_superuser and not user.is_superuser:
                raise credentials_exception

            return SchemaUser.model_validate(user)

        return func
